import traceback
from logging import Handler, LogRecord

from contek_tusk import Metric
from contek_tusk.logging.tusk_logging_config import TuskLoggingConfig

ORIGINAL_EXTRA_KEY = '_extra'


class TuskHandler(Handler):

    def __init__(
        self,
        metric: Metric,
        config: TuskLoggingConfig,
        level: int,
    ) -> None:
        super().__init__(level)
        self._metric = metric
        self._config = config

    def emit(self, record: LogRecord) -> None:
        config = self._config
        key_values = {}

        level = config.get_mapper()(record.levelno)
        key_values[config.get_level_column()] = level

        app = config.get_app_name()
        key_values[config.get_app_column()] = app

        logger_column = config.get_logger_column()
        if logger_column is not None:
            key_values[logger_column] = record.filename

        line_column = config.get_line_column()
        if line_column is not None:
            key_values[line_column] = record.lineno

        error_column = config.get_error_column()
        if error_column is not None:
            exc = record.exc_info
            error = exc[0].__name__ if exc else ''
            key_values[error_column] = error

        message_column = config.get_message_column()
        if message_column is not None:
            msg = record.msg
            key_values[message_column] = msg if msg is not None else ''

        stacktrace_column = config.get_stacktrace_column()
        if stacktrace_column is not None:
            stacktrace = ''
            if record.exc_info:
                lines = traceback.format_exception(*record.exc_info)
                stacktrace = ''.join(lines)
            key_values[stacktrace_column] = stacktrace

        context_columns = config.get_context_columns()
        if hasattr(record, ORIGINAL_EXTRA_KEY):
            extra = getattr(record, ORIGINAL_EXTRA_KEY)
            for (key, value) in extra.items():
                column = context_columns.get(key, key)
                key_values[column] = value

        self._metric.write(key_values)



from datetime import timedelta
from typing import Dict, Optional, Callable

from contek_tusk.app_name_resolver import APP_NAME_KEY
from contek_tusk.batching_config import BatchingConfig
from contek_tusk.logging.level_mapper import LevelMapper
from contek_tusk.table import Table


class TuskLoggingConfig:
    DEFAULT_BATCHING_CONFIG = BatchingConfig(timedelta(seconds=10), 50000)
    DEFAULT_APP_COLUMN = APP_NAME_KEY
    DEFAULT_LEVEL_COLUMN = 'level'
    DEFAULT_LOGGER_COLUMN = 'logger'
    DEFAULT_LINE_COLUMN = 'line'
    DEFAULT_ERROR_COLUMN = 'error'
    DEFAULT_MESSAGE_COLUMN = 'message'
    DEFAULT_STACKTRACE_COLUMN = 'stacktrace'

    def __init__(
        self,
        app_name: str,
        table: Table,
        batching_config: BatchingConfig = DEFAULT_BATCHING_CONFIG,
        app_column: str = DEFAULT_APP_COLUMN,
        level_column: str = DEFAULT_LEVEL_COLUMN,
        logger_column: Optional[str] = DEFAULT_LOGGER_COLUMN,
        line_column: Optional[str] = DEFAULT_LINE_COLUMN,
        error_column: Optional[str] = DEFAULT_ERROR_COLUMN,
        message_column: Optional[str] = DEFAULT_MESSAGE_COLUMN,
        stacktrace_column: Optional[str] = DEFAULT_STACKTRACE_COLUMN,
        context_columns: Optional[Dict[str, str]] = None,
        mapper: Optional[Callable[[int], int]] = None,
    ) -> None:
        self._app_name = app_name
        self._table = table
        self._batching_config = batching_config
        self._app_column = app_column
        self._level_column = level_column
        self._logger_column = logger_column
        self._line_column = line_column
        self._error_column = error_column
        self._message_column = message_column
        self._stacktrace_column = stacktrace_column
        self._context_columns = {} if context_columns is None else context_columns
        self._mapper = LevelMapper() if mapper is None else mapper

    def get_app_name(self) -> str:
        return self._app_name

    def get_table(self) -> Table:
        return self._table

    def get_batching_config(self) -> BatchingConfig:
        return self._batching_config

    def get_app_column(self) -> str:
        return self._app_column

    def get_level_column(self) -> str:
        return self._level_column

    def get_logger_column(self) -> Optional[str]:
        return self._logger_column

    def get_line_column(self) -> Optional[str]:
        return self._line_column

    def get_error_column(self) -> Optional[str]:
        return self._error_column

    def get_message_column(self) -> Optional[str]:
        return self._message_column

    def get_stacktrace_column(self) -> Optional[str]:
        return self._stacktrace_column

    def get_context_columns(self) -> Dict[str, str]:
        return self._context_columns

    def get_mapper(self) -> Callable[[int], int]:
        return self._mapper



from threading import Lock
from typing import Optional

from contek_tusk.schema import Schema
from contek_tusk.schema_provider import SchemaProvider


class CountColumnCache:

    def __init__(
        self,
        schema_provider: SchemaProvider,
        count_column: Optional[str],
    ) -> None:
        self._schema_provider = schema_provider
        self._cache: Optional[str] = count_column
        self._lock: Lock = Lock()

    def get(self) -> Optional[str]:
        self._lock.acquire()

        try:
            if self._cache is not None:
                return self._cache

            schema = self._schema_provider.get_schema()
            if schema is None:
                return None

            self._cache = CountColumnCache._get_first_int_column(schema)
            return self._cache
        finally:
            self._lock.release()

    @staticmethod
    def _get_first_int_column(schema: Schema) -> Optional[str]:
        for (column_name, column_type) in schema.get_columns().items():
            if column_type.startswith('UInt') or column_type.startswith('Int'):
                return column_name
        return None



from threading import Lock
from typing import Dict

from contek_tusk.counter.count_column_cache import CountColumnCache
from contek_tusk.counter.count_details import CountDetails
from contek_tusk.metric import Metric


class CountBuffer:

    def __init__(
        self,
        metric: Metric,
        count_column_cache: CountColumnCache,
    ) -> None:
        self._metric = metric
        self._count_column_cache = count_column_cache
        self._counts: Dict[str, CountDetails] = {}
        self._lock = Lock()

    def count(self, tags: Dict[str, str], n: int = 1) -> None:
        key = CountBuffer._to_tags_key(tags)
        self._lock.acquire()
        try:
            count = self._counts.get(key)
            if count is None:
                count = CountDetails(tags)
                self._counts[key] = count
            count.increment_unsafe(n)
        finally:
            self._lock.release()

    def flush(self) -> bool:
        count_column = self._count_column_cache.get()
        if count_column is None:
            return False

        self._lock.acquire()
        try:
            while len(self._counts) > 0:
                (_, count) = self._counts.popitem()
                key_values: Dict[str, any] = count.get_tags()
                key_values[count_column] = count.get_count_unsafe()
                self._metric.write(key_values)
        finally:
            self._lock.release()

    @staticmethod
    def _to_tags_key(tags: Dict[str, str]) -> str:
        return '&'.join(["{}={}".format(k, v) for (k, v) in tags.items()])



from typing import Dict, Optional


class Schema:

    def __init__(self, columns: Dict[str, str]) -> None:
        self._columns = columns

    def get_column_type(self, column_name: str) -> Optional[str]:
        return self._columns.get(column_name)

    def get_columns(self) -> Dict[str, str]:
        return self._columns



from threading import Lock
from typing import Optional

import contek_tusk as tusk
from contek_tusk.schema import Schema
from contek_tusk.table import Table


class SchemaProvider:

    def __init__(self, table: Table, client_name: str) -> None:
        self._table = table
        self._client_name = client_name
        self._cache: Optional[Schema] = None
        self._lock: Lock = Lock()

    def get_schema(self) -> Optional[Schema]:
        self._lock.acquire()

        try:
            if self._cache is not None:
                return self._cache

            self._cache = self._fetch()
            return self._cache
        finally:
            self._lock.release()

    def _fetch(self) -> Optional[Schema]:
        client = tusk.get_client(self._client_name)
        if client is None:
            return None

        return client.describe(self._table)



import logging
from logging import INFO
from typing import Union, Dict, Optional

import contek_tusk
from contek_tusk.app_name_resolver import get_app_name
from contek_tusk.logging.tusk_handler import TuskHandler, ORIGINAL_EXTRA_KEY
from contek_tusk.logging.tusk_logging_config import TuskLoggingConfig
from contek_tusk.metric_client import DEFAULT_CLIENT
from contek_tusk.table import Table

_handlers: Dict[str, TuskHandler] = {}


def init(
    table: Union[Table, str],
    app_name: Optional[str] = None,
    client_name: str = DEFAULT_CLIENT,
    level: int = INFO,
) -> None:
    if type(table) is str:
        table = Table.from_str(table)
    if app_name is None:
        app_name = get_app_name()
    init_for_config(
        TuskLoggingConfig(
            app_name=app_name,
            table=table,
        ),
        client_name,
        level,
    )


def init_for_config(
    config: TuskLoggingConfig,
    client_name: str = DEFAULT_CLIENT,
    level: int = INFO,
) -> None:
    metric = contek_tusk.metric(
        config.get_table(),
        config.get_batching_config(),
        client_name,
    )
    _patch_make_record_extra()
    handler = TuskHandler(metric, config, level)
    logger = logging.getLogger()
    logger.addHandler(handler)
    global _handlers
    _handlers[client_name] = handler


def _patch_make_record_extra():
    make_record = logging.Logger.makeRecord
    var_names = make_record.__code__.co_varnames

    def make_record_with_extra(*args, **kwargs):
        kwargs.update(zip(var_names, args))
        record = make_record(**kwargs)
        extra = kwargs.get('extra')
        if extra is not None:
            setattr(record, ORIGINAL_EXTRA_KEY, extra)
        return record

    logging.Logger.makeRecord = make_record_with_extra



from pandas import DataFrame

from contek_tusk.table import Table


class MetricData:

    def __init__(self, table: Table, df: DataFrame) -> None:
        self._table = table
        self._df = df

    def get_table(self) -> Table:
        return self._table

    def get_data_frame(self) -> DataFrame:
        return self._df



from __future__ import annotations

from typing import Optional, Dict

DATABASE = 'database'
TABLE = 'table'


class Table:

    def __init__(
        self,
        database: Optional[str],
        table_name: str,
        time_column: Optional[str] = None,
    ) -> None:
        self._database = database
        self._table_name = table_name
        self._time_column = time_column

    @classmethod
    def from_str(cls, str_value: str) -> Table:
        split = str_value.split('.')
        if len(split) == 2:
            return cls(split[0], split[1])
        if len(split) == 1:
            return cls(None, split[0])
        raise ValueError(str_value)

    @classmethod
    def from_dict(cls, key_values: Dict[str, str]) -> Table:
        database = key_values.get(DATABASE)
        table = key_values.get(TABLE)
        if table is None:
            raise ValueError('Table not specified')
        return cls(database, table)

    def get_full_name(self) -> str:
        if self._database is None:
            return self._table_name

        return f"{self._database}.{self._table_name}"

    def get_time_column(self) -> Optional[str]:
        return self._time_column



from typing import List, Optional

from pandas import DataFrame

from contek_tusk.entry_row import EntryRow
from contek_tusk.metric_data import MetricData
from contek_tusk.schema_provider import SchemaProvider
from contek_tusk.table import Table


class MetricFormatter:

    def __init__(
        self,
        table: Table,
        schema_provider: SchemaProvider,
    ) -> None:
        self._table = table
        self._schema_provider = schema_provider

    def format(self, rows: List[EntryRow]) -> Optional[MetricData]:
        schema = self._schema_provider.get_schema()
        if schema is None:
            return None

        if len(rows) == 0:
            return None

        records = list(map(lambda r: r.get_key_values(), rows))
        df = DataFrame(records, dtype=object)
        return MetricData(self._table, df)



from threading import Lock
from typing import Optional, Dict

from contek_tusk.env_utils import get_env_tag
from contek_tusk.schema import Schema
from contek_tusk.schema_provider import SchemaProvider


class EnvTagsCache:

    def __init__(self, schema_provider: SchemaProvider) -> None:
        self._schema_provider = schema_provider
        self._cache: Optional[Dict[str, str]] = None
        self._lock: Lock = Lock()

    def get(self) -> Optional[Dict[str, str]]:
        self._lock.acquire()

        try:
            if self._cache is not None:
                return self._cache

            schema = self._schema_provider.get_schema()
            if schema is None:
                return None

            self._cache = EnvTagsCache._get_tags_from_env(schema)
            return self._cache
        finally:
            self._lock.release()

    @staticmethod
    def _get_tags_from_env(schema: Schema) -> Dict[str, str]:
        result: Dict[str, str] = {}
        for (column_name, column_type) in schema.get_columns().items():
            if column_type != 'String':
                continue
            env_value = get_env_tag(column_name)
            if env_value is not None:
                result[column_name] = env_value
        return result



from typing import Dict, Optional


class EntryRow:

    def __init__(self, key_values: Dict[str, any]) -> None:
        self._key_values = key_values

    def get_key_values(self) -> Dict[str, any]:
        return self._key_values

    def get_value(self, key: str) -> Optional[any]:
        return self._key_values[key]



from datetime import timedelta
from typing import Union, Dict, Optional

from contek_tusk.app_name_resolver import get_app_name
from contek_tusk.heartbeat.tuck_heartbeat_client import TuskHeartbeatClient
from contek_tusk.heartbeat.tuck_heartbeat_config import TuskHeartbeatConfig
from contek_tusk.metric_client import DEFAULT_CLIENT
from contek_tusk.table import Table

DEFAULT_HEARTBEAT_PERIOD = timedelta(seconds=30)
DEFAULT_TIMEOUT = timedelta(seconds=30)

_heartbeat_clients: Dict[str, TuskHeartbeatClient] = {}


def init(
    table: Union[Table, str],
    app_name: Optional[str] = None,
    heartbeat_period: Union[timedelta, int] = DEFAULT_HEARTBEAT_PERIOD,
    timeout: Union[timedelta, int] = DEFAULT_TIMEOUT,
    client_name: str = DEFAULT_CLIENT,
) -> None:
    if type(table) is str:
        table = Table.from_str(table)
    if type(heartbeat_period) is int:
        heartbeat_period = timedelta(seconds=heartbeat_period)
    if type(timeout) is int:
        timeout = timedelta(seconds=timeout)
    if app_name is None:
        app_name = get_app_name()
    init_for_config(
        TuskHeartbeatConfig(
            app_name=app_name,
            table=table,
            heartbeat_period=heartbeat_period,
            timeout=timeout,
        ),
        client_name,
    )


def init_for_config(
    config: TuskHeartbeatConfig,
    client_name: str = DEFAULT_CLIENT,
) -> None:
    client = TuskHeartbeatClient.create(config, client_name)
    client.start()
    global _heartbeat_clients
    _heartbeat_clients[client_name] = client


def beat(
    task: str,
    heartbeat_period: Union[timedelta, int],
    client_name: str = DEFAULT_CLIENT,
) -> None:
    client = _get_client(client_name)
    if client is not None:
        client.beat(task, heartbeat_period)


def _get_client(client_name: str) -> Optional[TuskHeartbeatClient]:
    global _heartbeat_clients
    client = _heartbeat_clients.get(client_name)
    return client



from __future__ import annotations

import atexit
from datetime import timedelta
from threading import RLock, Timer
from typing import Dict, Union, Optional

from contek_tusk.batching_config import BatchingConfig
from contek_tusk.counter.count_buffer import CountBuffer
from contek_tusk.counter.count_column_cache import CountColumnCache
from contek_tusk.metric import Metric
from contek_tusk.metric_client import DEFAULT_CLIENT
from contek_tusk.table import Table


class TuskCounter:
    DEFAULT_COUNT_PERIOD = timedelta(seconds=15)
    DEFAULT_BATCHING_CONFIG = BatchingConfig.default()

    def __init__(self, buffer: CountBuffer, count_period: timedelta) -> None:
        self._buffer = buffer
        self._count_period = count_period
        self._lock = RLock()
        self._timer: Optional[Timer] = None
        atexit.register(self._flush)

    @classmethod
    def counter(
        cls,
        table: Union[Table, str],
        count_column: Optional[str] = None,
        count_period: Union[timedelta, int] = DEFAULT_COUNT_PERIOD,
        batching_config: BatchingConfig = DEFAULT_BATCHING_CONFIG,
        client_name: str = DEFAULT_CLIENT,
    ) -> TuskCounter:
        metric = Metric.metric(table, batching_config, client_name)
        schema_provider = metric.get_schema_provider()
        count_column_cache = CountColumnCache(schema_provider, count_column)
        buffer = CountBuffer(metric, count_column_cache)
        if type(count_period) is int:
            count_period = timedelta(seconds=count_period)
        return cls(buffer, count_period)

    def count(self, tags: Dict[str, str], increment: int = 1) -> None:
        self._buffer.count(tags, increment)
        self._schedule_if_idle()

    def _schedule_if_idle(self) -> None:
        self._lock.acquire()
        try:
            timer = self._timer
            if timer is not None and timer.is_alive():
                return
            self._schedule()
        finally:
            self._lock.release()

    def _flush_and_schedule(self) -> None:
        updated = self._flush()
        if not updated:
            return
        self._schedule()

    def _schedule(self) -> None:
        self._lock.acquire()
        try:
            timer = Timer(
                self._count_period.total_seconds(),
                self._flush_and_schedule,
            )
            timer.daemon = True
            timer.start()
            self._timer = timer
        finally:
            self._lock.release()

    def _flush(self) -> bool:
        return self._buffer.flush()



import datetime
from typing import Dict, Optional

from contek_tusk.entry_row import EntryRow
from contek_tusk.schema import Schema
from contek_tusk.schema_provider import SchemaProvider


class EntryInputNormalizer:

    def __init__(self, schema_provider: SchemaProvider) -> None:
        self._schema_provider = schema_provider

    def normalize(self, key_values: Dict[str, any]) -> Optional[EntryRow]:
        schema = self._schema_provider.get_schema()
        if schema is None:
            return None

        result: Dict[str, any] = {}
        for (key, value) in key_values.items():
            normalized = EntryInputNormalizer._normalize_for_column(
                value,
                key,
                schema,
            )
            if normalized is not None:
                result[key] = normalized

        for (column_name, column_type) in schema.get_columns().items():
            if column_name not in result:
                fill = EntryInputNormalizer._fill_for_type(column_type)
                result[column_name] = fill

        return EntryRow(result)

    @staticmethod
    def _normalize_for_column(
        value: any,
        column_name: str,
        schema: Schema,
    ) -> any:
        if schema is None:
            return value

        column_type = schema.get_column_type(column_name)
        if type is None:
            raise RuntimeError(
                f"Column \"{column_name}\" is not defined in the schema")

        return EntryInputNormalizer._normalize_for_type(value, column_type)

    @staticmethod
    def _normalize_for_type(value: any, column_type: str) -> any:
        value_type = type(value)
        if column_type == 'String':
            if value_type is not str:
                return str(value)

        if column_type.startswith('DateTime'):
            if value_type == float:
                return datetime.datetime.utcfromtimestamp(value)

        if column_type.startswith('Int') or column_type.startswith('UInt'):
            if value_type == str or value_type == float:
                return int(value)

        if column_type.startswith('Float'):
            if value_type == str or value_type == int:
                return float(value)

        return value

    @staticmethod
    def _fill_for_type(column_type: str) -> any:
        if column_type == 'String':
            return ''

        if column_type.startswith('DateTime'):
            return datetime.datetime.utcfromtimestamp(0)

        if column_type.startswith('Int') or column_type.startswith('UInt'):
            return int(0)

        if column_type.startswith('Float'):
            return float(0)

        raise ValueError(column_type)



from __future__ import annotations

import logging
from typing import Optional, Dict

from clickhouse_driver import Client
from clickhouse_driver.errors import Error

from contek_tusk.metric_data import MetricData
from contek_tusk.schema import Schema
from contek_tusk.table import Table

DEFAULT_CLIENT = 'default'
logger = logging.getLogger(__name__)


class MetricClient:

    def __init__(self, client: Client):
        self._client = client

    @classmethod
    def create(
        cls,
        host: str,
        user: str,
        password: str,
        **kwargs,
    ) -> MetricClient:
        client = Client(
            host=host,
            user=user,
            password=password,
            settings={"use_numpy": True},
            **kwargs,
        )
        return cls(client)

    def write(self, data: MetricData) -> None:
        full_table_name = data.get_table().get_full_name()
        query = f"INSERT INTO {full_table_name} VALUES"

        df = data.get_data_frame()
        rows = df.shape[0]
        if rows < 1:
            return
        try:
            self._client.insert_dataframe(query, df)
        except Error:
            logger.exception(
                f"Failed to flush metric data into table \"{data.get_table().get_full_name()}\"."
            )

    def describe(self, table: Table) -> Optional[Schema]:
        full_table_name = table.get_full_name()
        query = f"DESCRIBE TABLE {full_table_name}"
        try:
            df = self._client.query_dataframe(query)
        except Error:
            logger.exception(
                f"Failed to describe table \"{table.get_full_name()}\".")
            return None
        result: Dict[str, str] = {}
        for (index, row) in df.iterrows():
            column_name = row.get('name')
            column_type = row.get('type')
            result[column_name] = column_type
        return Schema(result)



import os
from typing import Dict

import yaml

import contek_tusk as tusk
import contek_tusk.heartbeat as tusk_heartbeat
import contek_tusk.logging as tusk_logging
from contek_tusk.app_name_resolver import set_app_name
from contek_tusk.table import Table

APP = 'app'
HEARTBEAT = 'heartbeat'
LOGGING = 'logging'


def for_yaml_file(file: str) -> None:
    with open(os.path.expanduser(file), 'r') as file:
        key_values = yaml.load(file, Loader=yaml.FullLoader)
    for_dict(key_values)


def for_dict(key_values: Dict[str, any]) -> None:
    app = key_values.get(APP)
    if type(app) is str:
        set_app_name(app)

    tusk.init(**key_values)

    heartbeat = key_values.get(HEARTBEAT)
    if type(heartbeat) is Dict:
        tusk_heartbeat.init(Table.from_dict(heartbeat))

    logging = key_values.get(LOGGING)
    if type(logging) is Dict:
        tusk_logging.init(Table.from_dict(logging))



from __future__ import annotations

from datetime import timedelta


class BatchingConfig:

    def __init__(self, duration: timedelta, max_size: int) -> None:
        self._duration = duration
        self._max_size = max_size

    @classmethod
    def disabled(cls) -> BatchingConfig:
        return cls(timedelta(seconds=0), 0)

    @classmethod
    def default(cls) -> BatchingConfig:
        return cls(timedelta(seconds=60), 10000)

    @classmethod
    def unlimited(cls, duration: timedelta) -> BatchingConfig:
        return cls(duration, 0)

    def is_immediate(self) -> bool:
        return self.get_period() == 0

    def get_period(self) -> timedelta:
        return self._duration

    def get_max_size(self) -> int:
        return self._max_size



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='contek-tusk',
    version='1.5',
    description='Tusk for Clickhouse metric writing',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    py_modules=["contek_tusk"],
    url='https://github.com/contek-io/contek-tusk',
    author='contek_bjy',
    author_email='bjy@contek.io',
    license='MIT',
    install_requires=[
        'clickhouse-driver',
        'pandas',
        'pyyaml',
    ],
    zip_safe=True,
)



from typing import Optional

from contek_tusk.env_utils import get_env_tag

APP_NAME_KEY = 'app'

_app_name: Optional[str] = None


def set_app_name(app_name: str) -> None:
    global _app_name
    _app_name = app_name


def get_app_name() -> str:
    global _app_name
    if _app_name is None:
        _app_name = get_env_tag(APP_NAME_KEY)
        if _app_name is None:
            raise ValueError('Could not read app name from env')
    return _app_name
